local addon, shared = ...

local RELEASE=		100
local SILENT=		99
local TESTING=		20
local WDOGYIELD=	15
local FUNCTIONRUN=	10
local STATUSUPDT=	8
local INFORMATION=	5
local DETAIL=		1

local STRINGSUB = string.sub
local STRINGFORMAT = string.format
local STRINGUPPER = string.upper
local STRINGLOWER = string.lower
local STRINGFIND = string.find

---------------------
-- Variables
---------------------

_JABOB = {}
local JB = _JABOB

--local SLOT_UPDATES = {}
local SLOT_PULSE = {}
local SLOT_PULSE_ACTIVE = false
local sp_time = 0
local sp_inc = -0.1
local CDACTIVE = false
local CDSLOTS = {}
local prevCDUpdate = 0

local _ui = { bagslots = {}, sicontrols = {}, sbcontrols = {}, tooltip = { t1 = { lines = {} }, c1 = { lines = {} }, c2 = { lines = {} } } }
JB.ui = _ui
local _watchdogyield = {}
local _invbagsvisible = false

local MAXBAGSLOTS = 50
local MAXPLRBAGS = 10
local MAXBNKBAGS = 10

local NUMVAULTSLOTS = 1
local vaultslots = {}

local szBags = {
	["inventory"] = { 0,0,0,0,0,0,0,0,0,0 },
	["bank"] = { 0,0,0,0,0,0,0,0,0,0 }
}

local _SLOTSINUSE = { ["inventory"]={}, ["bank"]={}, ["vault"]={} }
local _SLOTID = {}
local data_libinvSLOT = {}

local _BAG_SI_ROWS = 0
local _BAG_SI_COLS = 0
local _BAG_SB_ROWS = 0
local _BAG_SB_COLS = 0

local _bagupdate = {
	["inventory"]={resize=false, arrange=false},
	["bank"]={resize=false, arrange=false},
}

local popupCur = ""

local coroDead = false
local UINEEDED = true
local SVLOADED = false
local CMDTOOLTIP = false

local CONFIG_INV = false
local CONFIG_BNK = false

_JABOB.SIU = _SLOTSINUSE
_JABOB.UI = _ui

local JB_MODE_BOX = 1
local JB_MODE_ROW = 2
local JB_MODE_BAG = 3

local JB_MODE_DESC = { "Columns: ", "One bag/row", "Adjacent bags" }

local JB_TOOLTIP_ALL = 1
local JB_TOOLTIP_RFT = 2
local JB_TOOLTIP_CUS = 3

---------------------
-- Functions
---------------------
local default_settings = {
	x=500, y = 500, nc = 20, si_iconsz = 32,
	bnkx=500, bnky=200, bnk_nc = 20, sb_iconsz = 32,
	bw = 1, sx=1, sy=1,
	ibw = 1, isx = 1, isy = 1,
	bbw = 1, bsx = 1, bsy = 1,
	version = "1.00.0001", debuglevel = SILENT, i_locked = false, b_locked = false, b_vault = true, num_vault = 1,
	i_mode = JB_MODE_BOX, b_mode = JB_MODE_BOX, tooltip = JB_TOOLTIP_CUS
	}

local debuglevel = default_settings.debuglevel

local function Debug(s,l)
	if (l or INFORMATION) >= debuglevel then
		print(s)
	end
end

local function WATCHDOGWAIT(x)
	local rmx = Inspect.System.Watchdog()
	while rmx < 0.05 do
		Debug(x,WDOGYIELD)
		coroutine.yield()
		rmx = Inspect.System.Watchdog()
	end
end

local function MergeTable(o,n)
	for k,v in pairs(n) do
		if type(v) == "table" then
			if o[k] == nil then
				o[k] = {}
			end
	 	 	if type(o[k]) == 'table' then
	 			MergeTable(o[k], n[k])
	 	 	end
		else
			if o[k] == nil then
				o[k] = v
			end
		end
	end
end

local function getRarityRGB(r)
	if r == "uncommon" then		return {r=0.00, g=0.80, b=0.00}
	elseif r == "quest" then	return {r=0.80, g=0.80, b=0.00}
	elseif r == "rare" then		return {r=0.15, g=0.51, b=0.99}
	elseif r == "relic" then	return {r=1.00, g=0.60, b=0.00}
	elseif r == "epic" then		return {r=0.69, g=0.28, b=1.00}
	elseif r == "sellable" then	return {r=0.53, g=0.53, b=0.53}
	elseif r == nil then		return {r=1.00, g=1.00, b=1.00}
	elseif r == "quest" then	return {r=1.00, g=0.96, b=0.00}
	elseif r == "transcended" then	return {r=1.00, g=0.00, b=0.00}
	elseif r == "schange" then	return {r=0.75, g=0.75, b=0.25}
	elseif r == "spos" then		return {r=0.00, g=0.75, b=0.00}
	elseif r == "sneg" then		return {r=0.75, g=0.00, b=0.00}
	else						return {r=0.00, g=0.00, b=0.00}
	end
end

--  "", "", "", "", "", "",
-- "", "energyMax", "energyRegen", "", "", "",
-- "manaMax", "manaRegen", "movement", "", "", "powerMax",
-- "powerRegen", "", "", "", "",
-- "", "", "", "stealth", "stealthDetect",
-- "", "", "xp", "", "", and "".

local _itemstats = {
	{ "armor", "Armor" },
	{ "strength", "Strength" },
	{ "intelligence", "Intelligence" },
	{ "wisdom", "Wisdom" },
	{ "dexterity", "Dexterity" },
	{ "endurance", "Endurance" },
	{ "block", "Block" },
	{ "parry", "Parry" },
	{ "dodge", "Dodge" },
	{ "powerSpell", "Spell Power" },
	{ "critSpell", "Spell Critical Hit" },
	{ "critPower", "Crit Power" },
	{ "powerAttack", "Attack Power" },
	{ "critAttack", "Physical Crit" },
	{ "hit", "Hit" },
	{ "valor", "Valor" },
	{ "vengeance", "Vengeance" },
	{ "toughness", "Toughness" },
	{ "deflect", "Deflect" },
	{ "resistAll", "Resist All" },
	{ "resistAir", "Air Resist" },
	{ "resistDeath", "Death Resist" },
	{ "resistEarth", "Earth Resist" },
	{ "resistFire", "Fire Resist" },
	{ "resistLife", "Life Resist" },
	{ "resistWater", "Water Resist" }
}

function JB.GetTTDetails(si)
	local _dtls = {}
	local _stats = {}
	local d = Inspect.Item.Detail(si)
	if d ~= nil then
		table.insert(_dtls, { t="NAME", i=d.name, r=d.rarity } )
		if d.bound == true then
			if d.bind == "account" then
				table.insert(_dtls, { t="INFO", i="Bound to Account" } )
			else
				table.insert(_dtls, { t="INFO", i="Soulbound" } )
			end
		else
			if bind == "equip" then
				table.insert(_dtls, { t="INFO", i="Bind on equip" } )
			end
		end
		if d.lootable then
			table.insert(_dtls, { t="INFO", i="Container" })
		else
			if d.category then
				table.insert(_dtls, { t="INFO", i=d.category:gsub(" %a", STRINGUPPER):gsub("%a", STRINGUPPER, 1) })
			end
		end

		if d.slots then
			table.insert(_dtls, { t="INFO", i=STRINGFORMAT("%d Slot Bag", d.slots)})
		end

		if d.damageMin ~= nil then
			table.insert(_dtls, { t="INFO", i=STRINGFORMAT("DPS: %.1f", ((d.damageMin+d.damageMax)/2)/d.damageDelay)})
			table.insert(_dtls, { t="INFO", i=STRINGFORMAT("Weapon Damage: %d - %d every %.1fs", d.damageMin, d.damageMax, d.damageDelay)})
		end
		if d.range then
			table.insert(_dtls, { t="INFO", i=STRINGFORMAT("Range: %dm", d.range)})
		end
		if d.stats then
			for k,v in ipairs(_itemstats) do
				if d.stats[v[1]] then
					table.insert(_dtls, { t="INFO", i=STRINGFORMAT("%s: %d", v[2], d.stats[v[1]])})
					_stats[v[1]] = d.stats[v[1]]
				else
					_stats[v[1]] = 0
				end
			end
		else
			for k,v in ipairs(_itemstats) do
				_stats[v[1]] = 0
			end
		end
		if d.statsRune then
			table.insert(_dtls, { t="NAME", i="Rune Stats", r="uncommon" })
			for k,v in ipairs(_itemstats) do
				if d.statsRune[v[1]] then
					table.insert(_dtls, { t="INFO", i=STRINGFORMAT("%s: %d", v[2], d.statsRune[v[1]])})
					_stats[v[1]] = _stats[v[1]] + d.statsRune[v[1]]
				end
			end
		end

		if d.requiredLevel then
			table.insert(_dtls, { t="INFO", i=STRINGFORMAT("Requires Level: %d", d.requiredLevel)})
		end

		if d.requiredCalling then
			table.insert(_dtls, { t="INFO", i=STRINGFORMAT("Calling: %s", d.requiredCalling:gsub(" %a", STRINGUPPER):gsub("%a", STRINGUPPER, 1))})
		end

		if d.sell then
			local sv = d.sell
			local p = math.floor(sv/10000)
			sv = sv - (p*10000)
			local g = math.floor(sv/100)
			local c = sv-(g*100)
			table.insert(_dtls, { t="NAME", i=STRINGFORMAT("SELL VALUE: %dp %dg %ds", p,g,c), r="relic"})

			if (d.stack or 0) > 1 then
				local sv = d.sell * d.stack
				local p = math.floor(sv/10000)
				sv = sv - (p*10000)
				local g = math.floor(sv/100)
				local c = sv-(g*100)
				table.insert(_dtls, { t="NAME", i=STRINGFORMAT("SELL STACK: %dp %dg %ds", p,g,c), r="relic"})
			end
		end
		return _dtls, _stats, d
	else
		return nil, nil, nil
	end
end

function JB.FillTTDetails(tt, dtl)
	local frw = 0
	local frh = 0

	for k,v in ipairs(dtl) do
		if v.t == "NAME" then
			rgb = getRarityRGB(v.r)
			tt.lines[k].info:SetFontColor(rgb.r, rgb.g, rgb.b)
		else
			tt.lines[k].info:SetFontColor(1,1,1)
		end
		tt.lines[k].info:SetText(v.i)
		frw = math.max(frw, tt.lines[k].info:GetWidth())
		frh = frh + tt.lines[k].info:GetHeight()
		tt.lines[k].info:SetVisible(true)
	end

	for x=#dtl+1, 40 do
		tt.lines[x].info:SetVisible(false)
	end

	tt.frame:SetHeight(frh)
	tt.frame:SetWidth(frw)
	tt.borders.t:SetWidth(frw-10)
	tt.borders.b:SetWidth(frw-10)
	tt.borders.l:SetHeight(frh-10)
	tt.borders.r:SetHeight(frh-10)
	tt.frame:SetVisible(true)
end

function JB.CompareStats(ttdtls, st_equip, st_new)
	local diffhdr = true
	for k,v in ipairs(_itemstats) do
		local kv = v[1]
		local d = st_equip[kv] - st_new[kv]
		if d ~= 0 then
			if diffhdr then
				table.insert(ttdtls, {t="NAME", i="Stat changes if replaced:", r="schange"})
				diffhdr = false
			end
			if d>0 then
				table.insert(ttdtls, {t="NAME", i=STRINGFORMAT("%s: +%s", v[2], d), r="spos"})
			else
				table.insert(ttdtls, {t="NAME", i=STRINGFORMAT("%s: -%s", v[2], (0-d)), r="sneg"})
			end
		end
	end
end

function JB.Showtooltip(si)
	local ttdetails, ttstats, iidetail
	local ttdetailsc1, ttstatsc1, iidetailc1
	local ttdetailsc2, ttstatsc2, iidetailc2
	local ncomptt = 0

	ttdetails, ttstats, iidetail = JB.GetTTDetails(si)

	if iidetail ~= nil then
		local dtlpart = {}
		if iidetail.category then
			for tkn in string.gmatch(iidetail.category, "[^%s]+") do
				table.insert(dtlpart, tkn)
			end
			if dtlpart[1] == "planar" then
				ncomptt = 1
				ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.fcs")
			elseif dtlpart[1] == "armor" then
				if dtlpart[2] == "accessory" then
					if dtlpart[3] == "ring" then
						ncomptt = 2
						ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.rn1")
						ttdetailsc2, ttstatsc2, iidetailc2 = JB.GetTTDetails("seqp.rn2")
					elseif dtlpart[3] == "trinket" then
						ncomptt = 1
						ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.tkt")
					elseif dtlpart[3] == "neck" then
						ncomptt = 1
						ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.nck")
					elseif dtlpart[3] == "seal" then
						ncomptt = 1
						ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.sel")
					elseif dtlpart[3] == nil then
						ncomptt = 1
						ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.syn")
					end
				elseif dtlpart[2] == "cape" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.cpe")
				elseif dtlpart[3] == "head" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.hlm")
				elseif dtlpart[3] == "shoulders" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.shl")
				elseif dtlpart[3] == "chest" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.chs")
				elseif dtlpart[3] == "hands" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.glv")
				elseif dtlpart[3] == "waist" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.blt")
				elseif dtlpart[3] == "legs" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.lgs")
				elseif dtlpart[3] == "feet" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.fet")
				end
			elseif dtlpart[1] == "weapon" then
				if dtlpart[2] == "ranged" then
					ncomptt = 1
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.rng")
				else
					ncomptt = 2
					ttdetailsc1, ttstatsc1, iidetailc1 = JB.GetTTDetails("seqp.hmn")
					ttdetailsc2, ttstatsc2, iidetailc2 = JB.GetTTDetails("seqp.hof")
				end
			end
		end

		local uiw = UIParent:GetWidth()
		local uih = UIParent:GetHeight()

		local ttx = UI.Native.TooltipAnchor:GetLeft()+(UI.Native.TooltipAnchor:GetWidth()/2)
		local tty = UI.Native.TooltipAnchor:GetTop()+(UI.Native.TooltipAnchor:GetHeight()/2)

		local tt_h, tt_v, cm_a, tta_p, tta_r, os_x

		if ttx > (uiw/2) then
			tt_h = "RIGHT"
			cm_a = "LEFT"
			os_x = -20
		else
			tt_h = "LEFT"
			cm_a = "RIGHT"
			os_x = 20
		end

		if tty > (uih/2) then
			tt_v = "BOTTOM"
		else
			tt_v = "TOP"
		end

		tta_p = tt_v..tt_h
		tta_r = tt_v..cm_a

		_ui.tooltip.t1.frame:ClearPoint("TOPLEFT")
		_ui.tooltip.t1.frame:ClearPoint("TOPRIGHT")
		_ui.tooltip.t1.frame:ClearPoint("BOTTOMRIGHT")
		_ui.tooltip.t1.frame:ClearPoint("BOTTOMLEFT")

		_ui.tooltip.t1.frame:SetPoint(tta_p, UI.Native.TooltipAnchor, tta_p, os_x,0)

		JB.FillTTDetails(_ui.tooltip.t1, ttdetails)

		_ui.tooltip.c1.frame:SetVisible(false)
		_ui.tooltip.c2.frame:SetVisible(false)

		if ncomptt > 0 then
			_ui.tooltip.c1.frame:ClearPoint("TOPRIGHT")
			_ui.tooltip.c1.frame:ClearPoint("BOTTOMRIGHT")
			_ui.tooltip.c1.frame:ClearPoint("TOPLEFT")
			_ui.tooltip.c1.frame:ClearPoint("BOTTOMLEFT")
			_ui.tooltip.c1.frame:SetPoint(tta_p, _ui.tooltip.t1.frame, tta_r, os_x,0)
			if ttdetailsc1 ~= nil then
				table.insert(ttdetailsc1, 1, {t="INFO", i="Currently Equipped:"})
				JB.CompareStats(ttdetailsc1, ttstats, ttstatsc1)
				JB.FillTTDetails(_ui.tooltip.c1, ttdetailsc1)
			end
			if ncomptt == 2 then
				_ui.tooltip.c2.frame:ClearPoint("TOPRIGHT")
				_ui.tooltip.c2.frame:ClearPoint("BOTTOMRIGHT")
				_ui.tooltip.c2.frame:ClearPoint("TOPLEFT")
				_ui.tooltip.c2.frame:ClearPoint("BOTTOMLEFT")
				_ui.tooltip.c2.frame:SetPoint(tta_p, _ui.tooltip.c1.frame, tta_r, os_x,0)
				if ttdetailsc2 ~= nil then
					table.insert(ttdetailsc2, 1, {t="INFO", i="Currently Equipped:"})
					JB.CompareStats(ttdetailsc2, ttstats, ttstatsc2)
					JB.FillTTDetails(_ui.tooltip.c2, ttdetailsc2)
				end
			end
		end
		
		local r = Library.LibInventory.GetItemInfo(iidetail.type)
		if r then
			local ix = 0
			for k,v in pairs(r) do
				ix = ix+1
				local cnt = 0
				for s,c in pairs(v) do
					cnt = cnt + c
				end
				
				if _ui.tooltip.t1.count.details[ix] == nil then
					local f = UI.CreateFrame("Text", STRINGFORMAT("_ui.tooltip.t1.count.details:%d",ix), _ui.tooltip.t1.count)
					f:SetFontSize(12)
					if ix == 1 then
						f:SetPoint("BOTTOMLEFT", _ui.tooltip.t1.count, "BOTTOMLEFT")
					else
						f:SetPoint("BOTTOMLEFT", _ui.tooltip.t1.count.details[ix-1], "TOPLEFT")
					end
					_ui.tooltip.t1.count.details[ix] = f
				end
				if k == "PLAYER" then
					_ui.tooltip.t1.count.details[ix]:SetText("You : "..cnt)
				else
					_ui.tooltip.t1.count.details[ix]:SetText(k.." : "..cnt)
				end
					
				_ui.tooltip.t1.count.details[ix]:SetVisible(true)
			end
			for x=ix+1, #_ui.tooltip.t1.count.details do
				_ui.tooltip.t1.count.details[x]:SetVisible(false)
			end
			if ix > 0 then
				_ui.tooltip.t1.count:SetHeight(ix * _ui.tooltip.t1.count.details[1]:GetHeight())
				_ui.tooltip.t1.count:SetVisible(true)
			else
				_ui.tooltip.t1.count:SetVisible(false)
			end
		end
	end
end

function JB.doResizeBag(st)
	local siname
	local h,w
	local sx, sy, sz, bw
	local trl, trr, trt, trb
	local frame, mode, bagcnt
	local nr, nc
	
	if st == "inventory" then
		sx = JABOB_Settings.isx
		sy = JABOB_Settings.isy
		sz = JABOB_Settings.si_iconsz
		frame = _ui.playerinv
		bw = JABOB_Settings.ibw
		mode = JABOB_Settings.i_mode
		nc = _BAG_SI_COLS
		bagcnt = MAXPLRBAGS
		siname = Utility.Item.Slot.Inventory
		nr = _BAG_SI_ROWS
	elseif st == "bank" then
		sx = JABOB_Settings.bsx
		sy = JABOB_Settings.bsy
		sz = JABOB_Settings.sb_iconsz
		frame = _ui.playerbnk
		bw = JABOB_Settings.bbw
		mode = JABOB_Settings.b_mode
		nc = _BAG_SB_COLS
		bagcnt = MAXBNKBAGS
		siname = Utility.Item.Slot.Bank
		nr = _BAG_SB_ROWS
	end

	trl, trt, trr, trb = frame:GetTrimDimensions()
	
	for x=1,bagcnt do
		WATCHDOGWAIT("ResizeBag:"..x)
		for y=1,szBags[st][x] do
			local si = siname(x,y)
			_ui.bagslots[si].container:SetWidth(sz+bw+bw)
			_ui.bagslots[si].container:SetHeight(sz+bw+bw)
			_ui.bagslots[si].frame:SetHeight(sz+bw+bw)
			_ui.bagslots[si].frame:SetWidth(sz+bw+bw)
			_ui.bagslots[si].icon:SetHeight(sz)
			_ui.bagslots[si].icon:SetHeight(sz)
			_ui.bagslots[si].icon:SetWidth(sz)
		end
	end
	
	if JABOB_Settings.b_vault and st == "bank" then
		for si,v in pairs(Library.LibInventory.REFSLOTS["vault"]) do
			WATCHDOGWAIT("ResizeBag:"..si)
			_ui.bagslots[si].container:SetWidth(sz+bw+bw)
			_ui.bagslots[si].container:SetHeight(sz+bw+bw)
			_ui.bagslots[si].frame:SetHeight(sz+bw+bw)
			_ui.bagslots[si].frame:SetWidth(sz+bw+bw)
			_ui.bagslots[si].icon:SetHeight(sz)
			_ui.bagslots[si].icon:SetHeight(sz)
			_ui.bagslots[si].icon:SetWidth(sz)
		end
	end
	
	local bagslots = 0

	h = (nr*(sz+bw+bw))+((nr-1)*sy)+trt+trb
	if mode == JB_MODE_BOX or mode == JB_MODE_ROW then
		w = (nc*(sz+bw+bw))+((nc-1)*sx)+trl+trr
	elseif mode == JB_MODE_BAG then
		w = (nc*(sz+bw+bw))+((nc-1)*sx)+nc-4+trl+trr
	end

	frame:SetWidth(w)
	frame:SetHeight(h)
end

function JB.ResizeBag(st)
	_bagupdate[st].resize = true
end

function JB.ArrangeBagSlots(st)
	_bagupdate[st].arrange = true
end

function JB.doArrangeBagSlots(sid)

	WATCHDOGWAIT("ArrangeBagSlots:"..sid)

	if JABOB_Settings.b_vault then
		for k,_ in pairs(Inspect.Item.List()) do
			if Library.LibInventory.REFSLOTS["vault"][k] then
				local st, sn, ss = Utility.Item.Slot.Parse(k)
				if sn > JABOB_Settings.num_vault then
					JABOB_Settings.num_vault = sn
				end
			end
		end
	end

	for k, v in pairs(Library.LibInventory.REFSLOTS["vault"]) do
		local vt, vs, vi = Utility.Item.Slot.Parse(k)
		if vs <= JABOB_Settings.num_vault and JABOB_Settings.b_vault then
			_ui.bagslots[k].container:SetVisible(true)
		else
			_ui.bagslots[k].container:SetVisible(false)
		end
	end

	local row=0
	local cnt=0
	local ele={}
	local trl, trr, trt, trb
	local bagcnt, numcol, sx, sy, baseframe, mode, siname

	if sid == "inventory" then
		bagcnt = MAXPLRBAGS
		numcol = JABOB_Settings.nc
		sx = JABOB_Settings.isx
		sy = JABOB_Settings.isy
		baseframe = _ui.playerinv
		mode = JABOB_Settings.i_mode
		siname = Utility.Item.Slot.Inventory
	elseif sid == "bank" then
		bagcnt = MAXBNKBAGS
		numcol = JABOB_Settings.bnk_nc
		sx = JABOB_Settings.bsx
		sy = JABOB_Settings.bsy
		baseframe = _ui.playerbnk
		mode = JABOB_Settings.b_mode
		siname = Utility.Item.Slot.Bank
	end
	
	trl, trt, trr, trb = baseframe:GetTrimDimensions()

	if mode == JB_MODE_BOX then
		cnt = 0
		row = 1
		for x=1,bagcnt do
			WATCHDOGWAIT("ArrangeBag:"..x)
			for y=1,szBags[sid][x] do
				cnt = cnt+1
				local si = siname(x,y)
				if x==1 and y==1 then
					_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPLEFT", trl,trt)
					ele[row] = _ui.bagslots[si].container
				else
					if cnt > numcol then
						row=row+1
						cnt=1
						ele[row] = _ui.bagslots[si].container
						_ui.bagslots[si].container:SetPoint("TOPLEFT", ele[row-1], "BOTTOMLEFT", 0,sy)
					else
						_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPRIGHT", sx,0)
					end
				end
				baseframe = _ui.bagslots[si].container
			end
		end
		
		if JABOB_Settings.b_vault and sid == "bank" then
			for x=1,JABOB_Settings.num_vault do
				WATCHDOGWAIT("ArrangeBag:"..x)
				for y=1,32 do
					cnt=cnt+1
					local si = Utility.Item.Slot.Vault(x,y)
					if cnt > numcol then
						row=row+1
						cnt=1
						ele[row] = _ui.bagslots[si].container
						_ui.bagslots[si].container:SetPoint("TOPLEFT", ele[row-1], "BOTTOMLEFT", 0,sy)
					else
						_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPRIGHT", sx,0)
					end
					baseframe = _ui.bagslots[si].container
				end
			end
		end
		
		if sid == "inventory" then
			_BAG_SI_ROWS = row
			_BAG_SI_COLS = numcol
		else
			_BAG_SB_ROWS = row
			_BAG_SB_COLS = numcol
		end		
	elseif mode == JB_MODE_ROW then
		cnt = 0
		row = 0
		for x=1,bagcnt do
			WATCHDOGWAIT("ArrangeBag:"..x)
			cnt = math.max(cnt, szBags[sid][x])
			for y=1,szBags[sid][x] do
				local si = siname(x,y)
				if y == 1 then
					row = row+1
					if x==1 then
						_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPLEFT", trl,trt)
					else
						_ui.bagslots[si].container:SetPoint("TOPLEFT", ele[row-1], "BOTTOMLEFT", 0,sy)
					end
					ele[row] = _ui.bagslots[si].container
				else
					_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPRIGHT", sx,0)
				end
				baseframe = _ui.bagslots[si].container
			end
		end
		
		if JABOB_Settings.b_vault and sid == "bank" then
			cnt = math.max(cnt, 32)
			for x=1,JABOB_Settings.num_vault do
				WATCHDOGWAIT("ArrangeBag:"..x)
				for y=1,32 do
					local si = Utility.Item.Slot.Vault(x,y)
					if y==1 then
						row=row+1
						ele[row] = _ui.bagslots[si].container
						_ui.bagslots[si].container:SetPoint("TOPLEFT", ele[row-1], "BOTTOMLEFT", 0,sy)
					else
						_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPRIGHT", sx,0)
					end
					baseframe = _ui.bagslots[si].container
				end
				
			end
		end
		
		if sid == "inventory" then
			_BAG_SI_ROWS = row
			_BAG_SI_COLS = cnt
		else
			_BAG_SB_ROWS = row
			_BAG_SB_COLS = cnt
		end		
	elseif mode == JB_MODE_BAG then
		local pr_col, pr_row
		local maxrows = 0
		local maxcols = 0
		for x=1,bagcnt do
			WATCHDOGWAIT("ArrangeBag:"..x)
			row=1
			for y=1,szBags[sid][x] do
				local si = siname(x,y)
				if y==1 then
					maxcols = maxcols+4
					if x==1 then
						_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPLEFT", trl,trt)
					else
						_ui.bagslots[si].container:SetPoint("TOPLEFT", pr_row, "TOPRIGHT", sx+4, 0)
					end
					pr_col = _ui.bagslots[si].container
				else
					if y%4==1 then
						_ui.bagslots[si].container:SetPoint("TOPLEFT", pr_col, "BOTTOMLEFT", 0, sy)
						pr_col = _ui.bagslots[si].container
						row=row+1
					else
						_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPRIGHT", sx, 0)
					end

					if y%4==0 and row==1 then
						pr_row = _ui.bagslots[si].container
					end
				end
				baseframe = _ui.bagslots[si].container
			end
			maxrows = math.max(maxrows, row)
		end
		
		if JABOB_Settings.b_vault and sid == "bank" then
			for x=1,JABOB_Settings.num_vault do
				WATCHDOGWAIT("ArrangeBag:"..x)
				row=1
				for y=1,32 do
					local si = Utility.Item.Slot.Vault(x,y)
					if y==1 then
						maxcols = maxcols+4
						_ui.bagslots[si].container:SetPoint("TOPLEFT", pr_row, "TOPRIGHT", sx+4, 0)
						pr_col = _ui.bagslots[si].container
					else
						if y%4==1 then
							_ui.bagslots[si].container:SetPoint("TOPLEFT", pr_col, "BOTTOMLEFT", 0, sy)
							pr_col = _ui.bagslots[si].container
							row=row+1
						else
							_ui.bagslots[si].container:SetPoint("TOPLEFT", baseframe, "TOPRIGHT", sx, 0)
						end

						if y%4==0 and row==1 then
							pr_row = _ui.bagslots[si].container
						end
					end
					baseframe = _ui.bagslots[si].container
				end
				maxrows = math.max(maxrows, row)
			end
		end
		
		if sid == "inventory" then
			_BAG_SI_ROWS = maxrows
			_BAG_SI_COLS = maxcols
		else
			_BAG_SB_ROWS = maxrows
			_BAG_SB_COLS = maxcols
		end		
	end
	
	JB.ResizeBag(sid)
	
	for x=1, bagcnt do
		WATCHDOGWAIT("ArrangeBag:"..x)
		for y=1,szBags[sid][x] do
			local si = siname(x,y)
			if _ui.bagslots[si] then
				_ui.bagslots[si].container:SetVisible(true)
			end
		end
		for y=szBags[sid][x]+1, MAXBAGSLOTS do
			local si = siname(x,y)
			if _ui.bagslots[si] then
				_ui.bagslots[si].container:SetVisible(false)
			end
		end
	end
end

function JB.procEvent_Addon_SavedVariables_Load_End()
	if JABOB_Settings == nil then JABOB_Settings = {} end
	MergeTable(JABOB_Settings, default_settings)
	_ui.playerinv:SetPoint("TOPLEFT", UIParent, "TOPLEFT", JABOB_Settings.x, JABOB_Settings.y)
	_ui.playerbnk:SetPoint("TOPLEFT", UIParent, "TOPLEFT", JABOB_Settings.bnkx, JABOB_Settings.bnky)
	if JABOB_Settings.i_locked then
		_ui.sicontrols.locktext:SetText("LOCKED")
	else
		_ui.sicontrols.locktext:SetText("Unlocked")
	end
	if JABOB_Settings.b_locked then
		_ui.sbcontrols.locktext:SetText("LOCKED")
	else
		_ui.sbcontrols.locktext:SetText("Unlocked")
	end
	if JABOB_Settings.i_mode == JB_MODE_BOX then
		_ui.sicontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.nc))
	else
		_ui.sicontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.i_mode])
	end
	if JABOB_Settings.b_mode == JB_MODE_BOX then
		_ui.sbcontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.bnk_nc))
	else
		_ui.sbcontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.b_mode])
	end
	_ui.sicontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.si_iconsz))
	_ui.sbcontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.sb_iconsz))

	if JABOB_Settings.b_vault then
		_ui.sbcontrols.vaulttext:SetText("Included")
	else
		_ui.sbcontrols.vaulttext:SetText("Excluded")
	end

	debuglevel = JABOB_Settings.debuglevel
end

function JB.BuildBagElement(si, pf)
	local f = {}
	
	f.container = UI.CreateFrame("Frame", "f.container:"..si, pf)
	f.container:SetVisible(false)
	f.container:SetLayer(5)
	f.container:SetBackgroundColor(0,0,0,0.5)
	
	f.frame = UI.CreateFrame("Texture", "f.frame:"..si, f.container)
	f.frame:SetVisible(false)
	f.frame:SetLayer(5)
	f.frame:SetPoint("TOPLEFT", f.container, "TOPLEFT")

	f.icon = UI.CreateFrame("Texture", "f.icon:"..si, f.container)
	f.icon:SetLayer(10)
	f.icon:SetVisible(false)
	f.icon:SetPoint("CENTER", f.container, "CENTER", 0,0)
	--f.icon:SetTexture(addon.identifier, "textures/bg_empty.png")

	f.icon:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		Command.Item.Standard.Left(si)
	end, "Event.UI.Input.Mouse.Left.Down")

	f.icon:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		local _luptime = Inspect.Time.Real()
		local t,h
		t,h = Inspect.Cursor()
		if h ~= nil then
			Command.Item.Move(h, si)
			Command.Cursor(nil)
		end
	end, "Event.UI.Input.Mouse.Left.Up")

	f.container:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		local t,h
		t,h = Inspect.Cursor()
		if h ~= nil then
			Command.Item.Move(h, si)
			Command.Cursor(nil)
		end
	end, "Event.UI.Input.Mouse.Left.Up")

	function f.icon:MouseIn(h, si)
		if (JABOB_Settings.tooltip == JB_TOOLTIP_ALL) or (JABOB_Settings.tooltip == JB_TOOLTIP_RFT) then
			local d = Inspect.Item.Detail(si)
			if d ~= nil then
				CMDTOOLTIP = d.id
			end
		end
		if (JABOB_Settings.tooltip == JB_TOOLTIP_ALL) or (JABOB_Settings.tooltip == JB_TOOLTIP_CUS) then
			local d = Inspect.Item.Detail(si)
			if d ~= nil then
				JB.Showtooltip(si)
				CMDTOOLTIP = d.id
			else
				_ui.tooltip.t1.frame:SetVisible(false)
				_ui.tooltip.c1.frame:SetVisible(false)
				_ui.tooltip.c2.frame:SetVisible(false)
			end
		end
		popupCur = si
	end

	f.icon:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
		Command.Item.Standard.Right(si)
		f.icon:MouseIn(nil, si)
	end, "Event.UI.Input.Mouse.Right.Click")

	f.icon:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
		f.icon:MouseIn(nil, si)
	end, "Event.UI.Input.Mouse.Cursor.In")

	f.icon:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(self, h)
		if popupCur == si then
			_ui.tooltip.t1.frame:SetVisible(false)
			_ui.tooltip.c1.frame:SetVisible(false)
			_ui.tooltip.c2.frame:SetVisible(false)
			popupCur = ""
			CMDTOOLTIP = nil
		end
	end, "Event.UI.Input.Mouse.Cursor.Out")

	f.text = UI.CreateFrame("Text", "f.text:"..si, f.container)
	f.text:SetFontSize(11)
	f.text:SetText("")
	f.text:SetLayer(20)
	f.text:SetPoint("BOTTOMRIGHT", f.frame, "BOTTOMRIGHT", 0,0)
	f.text:SetEffectGlow({strength=5, blurX=2, blurY=2})
	f.text:SetVisible(false)
	
	f.cd = UI.CreateFrame("Text", "f.cd:"..si, f.container)
	f.cd:SetFontSize(11)
	f.cd:SetText("")
	f.cd:SetLayer(20)
	f.cd:SetPoint("TOPLEFT", f.frame, "TOPLEFT", 0,0)
	f.cd:SetEffectGlow({strength=5, blurX=2, blurY=2})
	f.cd:SetVisible(false)
	
	return f
end

function JB.SearchBags(skey)
	local txt
	local search
	
	if skey == "inventory" then
		txt = _ui.sicontrols.search:GetText()
		search={["inventory"] = true}
	elseif skey == "bank" then
		search={["bank"] = true, ["vault"] = true}
		txt = _ui.sbcontrols.search:GetText()
	end
	
	if txt == "" then
		for k1, v1 in pairs(search) do
			for k,v in pairs(_SLOTSINUSE[k1]) do
				_ui.bagslots[k].container:SetAlpha(1)
			end
		end
	else
		local styp = STRINGSUB(txt,1,1)
		local stxt = STRINGLOWER(txt)
		local sfld = "name"

		if styp == "=" then
			stxt = STRINGLOWER(STRINGSUB(txt,2) or "")
			sfld = "category"
		end
		for k1, v1 in pairs(search) do
			for k,v in pairs(_SLOTSINUSE[k1]) do
				if STRINGFIND(STRINGLOWER(v[sfld] or ""), stxt) then
					_ui.bagslots[k].container:SetAlpha(1)
				else
					_ui.bagslots[k].container:SetAlpha(0.2)
				end
			end
		end
	end
end

function JB.BuildUI()
	_ui.context = UI.CreateContext(addon.identifier)

	WATCHDOGWAIT("BuildUI:1")
	_ui.playerinv = UI.CreateFrame("RiftWindow", "_ui.playerinv", _ui.context)
	_ui.playerinv:SetVisible(false)
	_ui.playerinv:SetTitle("JABOB [Player Inventory]")
	_ui.playerinv:SetLayer(0)

	local invborder = _ui.playerinv:GetBorder()

	invborder:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		if JABOB_Settings.i_locked == false then
			self.MouseDown = true
			local mouseData = Inspect.Mouse()
			self.sx = mouseData.x - _ui.playerinv:GetLeft()
			self.sy = mouseData.y - _ui.playerinv:GetTop()
		end
	end, "Event.UI.Input.Mouse.Left.Down")

	invborder:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			_ui.playerinv:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")

	invborder:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.MouseDown then
			self.MouseDown = false
		end
		JABOB_Settings.x = _ui.playerinv:GetLeft()
		JABOB_Settings.y = _ui.playerinv:GetTop()
	end, "Event.UI.Input.Mouse.Left.Up")

	WATCHDOGWAIT("BuildUI:2")

	_ui.sicontrols.close = UI.CreateFrame("RiftButton", "_ui.sicontrols.close", _ui.playerinv)
	_ui.sicontrols.close:SetSkin("close")
	_ui.sicontrols.close:SetPoint("TOPRIGHT", _ui.playerinv, "TOPRIGHT", -6,6)
	_ui.sicontrols.close:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		JB.CloseBags("inventory")
	end, "Event.UI.Input.Mouse.Left.Click")

	_ui.sicontrols.search = UI.CreateFrame("RiftTextfield", "_ui.sicontrols.search", _ui.playerinv)
	_ui.sicontrols.search:SetPoint("CENTERRIGHT", _ui.sicontrols.close, "CENTERLEFT", 0, 0)
	_ui.sicontrols.search:SetLayer(10)
	_ui.sicontrols.search:SetWidth(100)
	_ui.sicontrols.search:SetBackgroundColor(0,0,0,0.75)

	_ui.sicontrols.search:EventAttach(Event.UI.Textfield.Change, function(self, h)
		JB.SearchBags("inventory")
	end, "Event.UI.Textfield.Change")

	_ui.sicontrols.options = UI.CreateFrame("Texture", "_ui.sicontrols.lock", _ui.playerinv)
	_ui.sicontrols.options:SetPoint("TOPLEFT", _ui.playerinv, "TOPLEFT", 14,14)
	_ui.sicontrols.options:SetTexture(addon.identifier, "textures/imgcog.png")
	_ui.sicontrols.options:SetWidth(24)
	_ui.sicontrols.options:SetHeight(24)
	_ui.sicontrols.options:SetLayer(10)

	_ui.sicontrols.config = UI.CreateFrame("Texture", "_ui.sicontrols.config", _ui.playerinv)
	_ui.sicontrols.config:SetPoint("BOTTOMRIGHT", _ui.sicontrols.options, "CENTER", 0,0)
	_ui.sicontrols.configAnchor = "BOTTOMRIGHT"
	_ui.sicontrols.config:SetWidth(160)
	_ui.sicontrols.config:SetHeight(166)
	_ui.sicontrols.config:SetLayer(100)
	_ui.sicontrols.config:SetTexture("Rift", "ItemToolTip_I71.dds")
	_ui.sicontrols.config:SetVisible(false)

	_ui.sicontrols.options:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		local frx = _ui.sicontrols.options:GetLeft()
		local fry = _ui.sicontrols.options:GetTop()
		local sx = UIParent:GetWidth()/2
		local sy = UIParent:GetHeight()/2
		local anchor

		if fry > sy then
			anchor = "BOTTOM"
		else
			anchor = "TOP"
		end

		if frx <= sx then
			anchor = anchor.."LEFT"
		else
			anchor = anchor.."RIGHT"
		end

		if anchor ~= _ui.sicontrols.configAnchor then
			_ui.sicontrols.config:ClearPoint(_ui.sicontrols.configAnchor)
			_ui.sicontrols.configAnchor = anchor
			_ui.sicontrols.config:SetPoint(anchor, _ui.sicontrols.options, "CENTER", 0,0)
		end

		CONFIG_INV = true
		if JABOB_Settings.i_mode == JB_MODE_BOX then
			_ui.sicontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.nc))
		else
			_ui.sicontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.i_mode])
		end
		_ui.sicontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.si_iconsz))
		if JABOB_Settings.i_locked then
			_ui.sicontrols.locktext:SetText("LOCKED")
		else
			_ui.sicontrols.locktext:SetText("Unlocked")
		end

		_ui.sicontrols.config:SetVisible(true)
	end, "Event.UI.Input.Mouse.Left.Click")

	_ui.sicontrols.layout = UI.CreateFrame("Texture", "_ui.sicontrols.layout", _ui.sicontrols.config)
	_ui.sicontrols.layout:SetPoint("TOPLEFT", _ui.sicontrols.config, "TOPLEFT", 10,10)
	_ui.sicontrols.layout:SetTexture(addon.identifier, "textures/layout.png")
	_ui.sicontrols.layout:SetLayer(12)
	_ui.sicontrols.layout:SetHeight(32)
	_ui.sicontrols.layout:SetWidth(32)
	_ui.sicontrols.layout:SetVisible(true)

	_ui.sicontrols.coltext = UI.CreateFrame("Text", "_ui.sicontrols.coltext", _ui.sicontrols.layout)
	_ui.sicontrols.coltext:SetPoint("CENTERLEFT", _ui.sicontrols.layout, "CENTERRIGHT", 10,0)
	_ui.sicontrols.coltext:SetLayer(12)
	_ui.sicontrols.coltext:SetFontSize(12)

	_ui.sicontrols.layout:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.i_mode == JB_MODE_BOX then
			JABOB_Settings.nc = JABOB_Settings.nc +1
			_ui.sicontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.nc))
			JB.ArrangeBagSlots("inventory")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sicontrols.layout:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.i_mode == JB_MODE_BOX then
			if JABOB_Settings.nc > 5 then
				JABOB_Settings.nc = JABOB_Settings.nc -1
				_ui.sicontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.nc))
				JB.ArrangeBagSlots("inventory")
			end
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")

	_ui.sicontrols.layout:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		if JABOB_Settings.i_mode == JB_MODE_BOX then
			JABOB_Settings.i_mode = JB_MODE_ROW
			_ui.sicontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.i_mode])
		elseif JABOB_Settings.i_mode == JB_MODE_ROW then
			JABOB_Settings.i_mode = JB_MODE_BAG
			_ui.sicontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.i_mode])
		else
			JABOB_Settings.i_mode = JB_MODE_BOX
			_ui.sicontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.nc))
		end
		JB.ArrangeBagSlots("inventory")
	end, "Event.UI.Input.Mouse.Left.Click")

	_ui.sicontrols.iconsize = UI.CreateFrame("Texture", "_ui.sicontrols.iconsize", _ui.sicontrols.config)
	_ui.sicontrols.iconsize:SetPoint("TOPLEFT", _ui.sicontrols.layout, "BOTTOMLEFT", 0,6)
	_ui.sicontrols.iconsize:SetTexture(addon.identifier, "textures/iconsize.png")
	_ui.sicontrols.iconsize:SetLayer(12)
	_ui.sicontrols.iconsize:SetHeight(32)
	_ui.sicontrols.iconsize:SetWidth(32)
	_ui.sicontrols.iconsize:SetVisible(true)

	_ui.sicontrols.icontext = UI.CreateFrame("Text", "_ui.sicontrols.icontext", _ui.sicontrols.config)
	_ui.sicontrols.icontext:SetPoint("CENTERLEFT", _ui.sicontrols.iconsize, "CENTERRIGHT", 10,0)
	_ui.sicontrols.icontext:SetLayer(12)
	_ui.sicontrols.icontext:SetFontSize(12)

	_ui.sicontrols.iconsize:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		JABOB_Settings.si_iconsz = JABOB_Settings.si_iconsz +1
		_ui.sicontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.si_iconsz))
		JB.ResizeBag("inventory")
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sicontrols.iconsize:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		JABOB_Settings.si_iconsz = JABOB_Settings.si_iconsz -1
		_ui.sicontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.si_iconsz))
		JB.ResizeBag("inventory")
	end, "Event.UI.Input.Mouse.Wheel.Back")

	_ui.sicontrols.lock = UI.CreateFrame("Texture", "_ui.sicontrols.lock", _ui.sicontrols.config)
	_ui.sicontrols.lock:SetPoint("TOPLEFT", _ui.sicontrols.iconsize, "BOTTOMLEFT", 0,6)
	_ui.sicontrols.lock:SetTexture(addon.identifier, "textures/imglock.png")
	_ui.sicontrols.lock:SetLayer(12)
	_ui.sicontrols.lock:SetHeight(32)
	_ui.sicontrols.lock:SetWidth(32)
	_ui.sicontrols.lock:SetVisible(true)

	_ui.sicontrols.locktext = UI.CreateFrame("Text", "_ui.sicontrols.locktext", _ui.sicontrols.config)
	_ui.sicontrols.locktext:SetPoint("CENTERLEFT", _ui.sicontrols.lock, "CENTERRIGHT", 10,0)
	_ui.sicontrols.locktext:SetLayer(12)
	_ui.sicontrols.locktext:SetFontSize(12)

	_ui.sicontrols.lock:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		JABOB_Settings.i_locked = not JABOB_Settings.i_locked
		if JABOB_Settings.i_locked then
			_ui.sicontrols.locktext:SetText("LOCKED")
		else
			_ui.sicontrols.locktext:SetText("Unlocked")
		end
	end, "Event.UI.Input.Mouse.Left.Click")

	WATCHDOGWAIT("BuildUI:3")

	_ui.sicontrols.spacing_h = UI.CreateFrame("Texture", "_ui.sicontrols.spacing_h", _ui.sicontrols.config)
	_ui.sicontrols.spacing_h:SetPoint("BOTTOMCENTER", _ui.sicontrols.config, "BOTTOMCENTER", 0, -12)
	_ui.sicontrols.spacing_h:SetTexture(addon.identifier, "textures/space_h.png")
	_ui.sicontrols.spacing_h:SetLayer(12)
	_ui.sicontrols.spacing_h:SetWidth(32)
	_ui.sicontrols.spacing_h:SetHeight(32)
	_ui.sicontrols.spacing_h:SetVisible(true)

	_ui.sicontrols.spacing_h:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.isx <= 7 then
			JABOB_Settings.isx = JABOB_Settings.isx +1
			JB.ArrangeBagSlots("inventory")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sicontrols.spacing_h:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.isx >= 1 then
			JABOB_Settings.isx = JABOB_Settings.isx -1
			JB.ArrangeBagSlots("inventory")
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")
	
	_ui.sicontrols.border_w = UI.CreateFrame("Texture", "_ui.sicontrols.border_w", _ui.sicontrols.config)
	_ui.sicontrols.border_w:SetPoint("TOPRIGHT", _ui.sicontrols.spacing_h, "TOPLEFT", -10, 0)
	_ui.sicontrols.border_w:SetTexture(addon.identifier, "textures/border_w.png")
	_ui.sicontrols.border_w:SetLayer(12)
	_ui.sicontrols.border_w:SetWidth(32)
	_ui.sicontrols.border_w:SetHeight(32)
	_ui.sicontrols.border_w:SetVisible(true)

	_ui.sicontrols.border_w:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.ibw <= 7 then
			JABOB_Settings.ibw = JABOB_Settings.ibw +1
			JB.ResizeBag("inventory")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sicontrols.border_w:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.ibw >= 1 then
			JABOB_Settings.ibw = JABOB_Settings.ibw -1
			JB.ResizeBag("inventory")
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")
	
	_ui.sicontrols.spacing_v = UI.CreateFrame("Texture", "_ui.sicontrols.spacing_v", _ui.sicontrols.config)
	_ui.sicontrols.spacing_v:SetPoint("TOPLEFT", _ui.sicontrols.spacing_h, "TOPRIGHT", 10, 0)
	_ui.sicontrols.spacing_v:SetTexture(addon.identifier, "textures/space_v.png")
	_ui.sicontrols.spacing_v:SetLayer(12)
	_ui.sicontrols.spacing_v:SetWidth(32)
	_ui.sicontrols.spacing_v:SetHeight(32)
	_ui.sicontrols.spacing_v:SetVisible(true)

	_ui.sicontrols.spacing_v:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.isy <= 7 then
			JABOB_Settings.isy = JABOB_Settings.isy +1
			JB.ArrangeBagSlots("inventory")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sicontrols.spacing_v:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.isy >= 1 then
			JABOB_Settings.isy = JABOB_Settings.isy -1
			JB.ArrangeBagSlots("inventory")
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")
	
	for sk, sv in pairs(Library.LibInventory.REFSLOTS["inventory"]) do
		WATCHDOGWAIT(STRINGFORMAT("BuildUI:4:%s", sk))
		_ui.bagslots[sk] = JB.BuildBagElement(sk, _ui.playerinv)
	end
	
	WATCHDOGWAIT("BuildUI:5")

	_ui.playerbnk = UI.CreateFrame("RiftWindow", "_ui.playerbnk", _ui.context)
	_ui.playerbnk:SetVisible(false)
	_ui.playerbnk:SetTitle("JABOB [Player Bank]")
	_ui.playerbnk:SetLayer(0)

	local bnkborder = _ui.playerbnk:GetBorder()

	bnkborder:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		if JABOB_Settings.b_locked == false then
			self.MouseDown = true
			local mouseData = Inspect.Mouse()
			self.sx = mouseData.x - _ui.playerbnk:GetLeft()
			self.sy = mouseData.y - _ui.playerbnk:GetTop()
		end
	end, "Event.UI.Input.Mouse.Left.Down")

	bnkborder:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			_ui.playerbnk:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")

	bnkborder:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.MouseDown then
			self.MouseDown = false
		end
		JABOB_Settings.bnkx = _ui.playerbnk:GetLeft()
		JABOB_Settings.bnky = _ui.playerbnk:GetTop()
	end, "Event.UI.Input.Mouse.Left.Up")

	WATCHDOGWAIT("BuildUI:6")

	_ui.sbcontrols.close = UI.CreateFrame("RiftButton", "_ui.sbcontrols.close", _ui.playerbnk)
	_ui.sbcontrols.close:SetSkin("close")
	_ui.sbcontrols.close:SetPoint("TOPRIGHT", _ui.playerbnk, "TOPRIGHT", -6,6)

	_ui.sbcontrols.close:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		JB.CloseBags("bank")
	end, "Event.UI.Input.Mouse.Left.Click")

	_ui.sbcontrols.search = UI.CreateFrame("RiftTextfield", "_ui.sbcontrols.search", _ui.playerbnk)
	_ui.sbcontrols.search:SetPoint("CENTERRIGHT", _ui.sbcontrols.close, "CENTERLEFT", 0,0)
	_ui.sbcontrols.search:SetLayer(10)
	_ui.sbcontrols.search:SetWidth(100)
	_ui.sbcontrols.search:SetBackgroundColor(0,0,0,0.75)

	_ui.sbcontrols.search:EventAttach(Event.UI.Textfield.Change, function(self, h)
		JB.SearchBags("bank")
	end, "Event.UI.Textfield.Change")

	_ui.sbcontrols.options = UI.CreateFrame("Texture", "_ui.sbcontrols.lock", _ui.playerbnk)
	_ui.sbcontrols.options:SetPoint("TOPLEFT", _ui.playerbnk, "TOPLEFT", 14,14)
	_ui.sbcontrols.options:SetTexture(addon.identifier, "textures/imgcog.png")
	_ui.sbcontrols.options:SetWidth(24)
	_ui.sbcontrols.options:SetHeight(24)
	_ui.sbcontrols.options:SetLayer(10)

	_ui.sbcontrols.config = UI.CreateFrame("Texture", "_ui.sbcontrols.config", _ui.playerbnk)
	_ui.sbcontrols.config:SetPoint("BOTTOMRIGHT", _ui.sbcontrols.options, "BOTTOMRIGHT", 0,0)
	_ui.sbcontrols.configAnchor = "BOTTOMRIGHT"
	_ui.sbcontrols.config:SetWidth(160)
	_ui.sbcontrols.config:SetHeight(204)
	_ui.sbcontrols.config:SetTexture("Rift", "ItemToolTip_I71.dds")
	_ui.sbcontrols.config:SetVisible(false)
	_ui.sbcontrols.config:SetLayer(100)

	_ui.sbcontrols.options:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		local frx = _ui.sbcontrols.options:GetLeft()
		local fry = _ui.sbcontrols.options:GetTop()
		local sx = UIParent:GetWidth()/2
		local sy = UIParent:GetHeight()/2
		local anchor

		if fry > sy then
			anchor = "BOTTOM"
		else
			anchor = "TOP"
		end

		if frx <= sx then
			anchor = anchor.."LEFT"
		else
			anchor = anchor.."RIGHT"
		end

		if anchor ~= _ui.sbcontrols.configAnchor then
			_ui.sbcontrols.config:ClearPoint(_ui.sbcontrols.configAnchor)
			_ui.sbcontrols.configAnchor = anchor
			_ui.sbcontrols.config:SetPoint(anchor, _ui.sbcontrols.options, "CENTER", 0,0)
		end

		CONFIG_BNK = true
		if JABOB_Settings.b_mode == JB_MODE_BOX then
			_ui.sbcontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.bnk_nc))
		else
			_ui.sbcontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.b_mode])
		end
		_ui.sbcontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.sb_iconsz))
		if JABOB_Settings.b_locked then
			_ui.sbcontrols.locktext:SetText("LOCKED")
		else
			_ui.sbcontrols.locktext:SetText("Unlocked")
		end

		_ui.sbcontrols.config:SetVisible(true)
	end, "Event.UI.Input.Mouse.Left.Click")

	_ui.sbcontrols.layout = UI.CreateFrame("Texture", "_ui.sbcontrols.layout", _ui.sbcontrols.config)
	_ui.sbcontrols.layout:SetPoint("TOPLEFT", _ui.sbcontrols.config, "TOPLEFT", 10,10)
	_ui.sbcontrols.layout:SetTexture(addon.identifier, "textures/layout.png")
	_ui.sbcontrols.layout:SetLayer(12)
	_ui.sbcontrols.layout:SetHeight(32)
	_ui.sbcontrols.layout:SetWidth(32)
	_ui.sbcontrols.layout:SetVisible(true)

	_ui.sbcontrols.coltext = UI.CreateFrame("Text", "_ui.sbcontrols.coltext", _ui.sbcontrols.layout)
	_ui.sbcontrols.coltext:SetPoint("CENTERLEFT", _ui.sbcontrols.layout, "CENTERRIGHT", 10,0)
	_ui.sbcontrols.coltext:SetLayer(12)
	_ui.sbcontrols.coltext:SetFontSize(12)

	_ui.sbcontrols.layout:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.b_mode == JB_MODE_BOX then
			JABOB_Settings.bnk_nc = JABOB_Settings.bnk_nc +1
			_ui.sbcontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.bnk_nc))
			JB.ArrangeBagSlots("bank")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sbcontrols.layout:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.b_mode == JB_MODE_BOX then
			if JABOB_Settings.bnk_nc > 5 then
				JABOB_Settings.bnk_nc = JABOB_Settings.bnk_nc -1
				_ui.sbcontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.bnk_nc))
				JB.ArrangeBagSlots("bank")
			end
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")

	_ui.sbcontrols.layout:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		if JABOB_Settings.b_mode == JB_MODE_BOX then
			JABOB_Settings.b_mode = JB_MODE_ROW
			_ui.sbcontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.b_mode])
		elseif JABOB_Settings.b_mode == JB_MODE_ROW then
			JABOB_Settings.b_mode = JB_MODE_BAG
			_ui.sbcontrols.coltext:SetText(JB_MODE_DESC[JABOB_Settings.b_mode])
		else
			JABOB_Settings.b_mode = JB_MODE_BOX
			_ui.sbcontrols.coltext:SetText(STRINGFORMAT("Columns: %d", JABOB_Settings.bnk_nc))
		end
		JB.ArrangeBagSlots("bank")
	end, "Event.UI.Input.Mouse.Left.Click")

	_ui.sbcontrols.iconsize = UI.CreateFrame("Texture", "_ui.sbcontrols.iconsize", _ui.sbcontrols.config)
	_ui.sbcontrols.iconsize:SetPoint("TOPLEFT", _ui.sbcontrols.layout, "BOTTOMLEFT", 0,6)
	_ui.sbcontrols.iconsize:SetTexture(addon.identifier, "textures/iconsize.png")
	_ui.sbcontrols.iconsize:SetLayer(12)
	_ui.sbcontrols.iconsize:SetHeight(32)
	_ui.sbcontrols.iconsize:SetWidth(32)
	_ui.sbcontrols.iconsize:SetVisible(true)

	_ui.sbcontrols.icontext = UI.CreateFrame("Text", "_ui.sbcontrols.icontext", _ui.sbcontrols.config)
	_ui.sbcontrols.icontext:SetPoint("CENTERLEFT", _ui.sbcontrols.iconsize, "CENTERRIGHT", 10,0)
	_ui.sbcontrols.icontext:SetLayer(12)
	_ui.sbcontrols.icontext:SetFontSize(12)

	_ui.sbcontrols.iconsize:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		JABOB_Settings.sb_iconsz = JABOB_Settings.sb_iconsz +1
		_ui.sbcontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.sb_iconsz))
		JB.ResizeBag("bank")
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sbcontrols.iconsize:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		JABOB_Settings.sb_iconsz = JABOB_Settings.sb_iconsz -1
		_ui.sbcontrols.icontext:SetText(STRINGFORMAT("Icon size: %d", JABOB_Settings.sb_iconsz))
		JB.ResizeBag("bank")
	end, "Event.UI.Input.Mouse.Wheel.Back")

	_ui.sbcontrols.lock = UI.CreateFrame("Texture", "_ui.sbcontrols.lock", _ui.sbcontrols.config)
	_ui.sbcontrols.lock:SetPoint("TOPLEFT", _ui.sbcontrols.iconsize, "BOTTOMLEFT", 0,6)
	_ui.sbcontrols.lock:SetTexture(addon.identifier, "textures/imglock.png")
	_ui.sbcontrols.lock:SetLayer(12)
	_ui.sbcontrols.lock:SetHeight(32)
	_ui.sbcontrols.lock:SetWidth(32)
	_ui.sbcontrols.lock:SetVisible(true)

	_ui.sbcontrols.locktext = UI.CreateFrame("Text", "_ui.sbcontrols.locktext", _ui.sbcontrols.config)
	_ui.sbcontrols.locktext:SetPoint("CENTERLEFT", _ui.sbcontrols.lock, "CENTERRIGHT", 10,0)
	_ui.sbcontrols.locktext:SetLayer(12)
	_ui.sbcontrols.locktext:SetFontSize(12)

	_ui.sbcontrols.lock:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		JABOB_Settings.b_locked = not JABOB_Settings.b_locked
		if JABOB_Settings.b_locked then
			_ui.sbcontrols.locktext:SetText("LOCKED")
		else
			_ui.sbcontrols.locktext:SetText("Unlocked")
		end
	end, "Event.UI.Input.Mouse.Left.Click")

	WATCHDOGWAIT("BuildUI:5b")

	_ui.sbcontrols.vault = UI.CreateFrame("Texture", "_ui.sbcontrols.vault", _ui.sbcontrols.config)
	_ui.sbcontrols.vault:SetPoint("TOPLEFT", _ui.sbcontrols.lock, "BOTTOMLEFT", 0,6)
	_ui.sbcontrols.vault:SetTexture(addon.identifier, "textures/vault.png")
	_ui.sbcontrols.vault:SetLayer(12)
	_ui.sbcontrols.vault:SetHeight(32)
	_ui.sbcontrols.vault:SetWidth(32)
	_ui.sbcontrols.vault:SetVisible(true)

	_ui.sbcontrols.vaulttext = UI.CreateFrame("Text", "_ui.sbcontrols.vaulttext", _ui.sbcontrols.config)
	_ui.sbcontrols.vaulttext:SetPoint("CENTERLEFT", _ui.sbcontrols.vault, "CENTERRIGHT", 10,0)
	_ui.sbcontrols.vaulttext:SetLayer(12)
	_ui.sbcontrols.vaulttext:SetFontSize(12)
	
	_ui.sbcontrols.vault:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		JABOB_Settings.b_vault = not JABOB_Settings.b_vault
		if JABOB_Settings.b_vault then
			_ui.sbcontrols.vaulttext:SetText("Included")
		else
			_ui.sbcontrols.vaulttext:SetText("Excluded")
		end
		JB.ArrangeBagSlots("bank")
	end, "Event.UI.Input.Mouse.Left.Click")
	
	_ui.sbcontrols.spacing_h = UI.CreateFrame("Texture", "_ui.sbcontrols.spacing_h", _ui.sbcontrols.config)
	_ui.sbcontrols.spacing_h:SetPoint("BOTTOMCENTER", _ui.sbcontrols.config, "BOTTOMCENTER", 0, -12)
	_ui.sbcontrols.spacing_h:SetTexture(addon.identifier, "textures/space_h.png")
	_ui.sbcontrols.spacing_h:SetLayer(12)
	_ui.sbcontrols.spacing_h:SetWidth(32)
	_ui.sbcontrols.spacing_h:SetHeight(32)
	_ui.sbcontrols.spacing_h:SetVisible(true)

	_ui.sbcontrols.spacing_h:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.bsx <= 7 then
			JABOB_Settings.bsx = JABOB_Settings.bsx +1
			JB.ArrangeBagSlots("bank")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sbcontrols.spacing_h:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.bsx >= 1 then
			JABOB_Settings.bsx = JABOB_Settings.bsx -1
			JB.ArrangeBagSlots("bank")
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")
	
	_ui.sbcontrols.border_w = UI.CreateFrame("Texture", "_ui.sbcontrols.border_w", _ui.sbcontrols.config)
	_ui.sbcontrols.border_w:SetPoint("TOPRIGHT", _ui.sbcontrols.spacing_h, "TOPLEFT", -10, 0)
	_ui.sbcontrols.border_w:SetTexture(addon.identifier, "textures/border_w.png")
	_ui.sbcontrols.border_w:SetLayer(12)
	_ui.sbcontrols.border_w:SetWidth(32)
	_ui.sbcontrols.border_w:SetHeight(32)
	_ui.sbcontrols.border_w:SetVisible(true)

	_ui.sbcontrols.border_w:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.bbw <= 7 then
			JABOB_Settings.bbw = JABOB_Settings.bbw +1
			JB.ResizeBag("bank")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sbcontrols.border_w:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.bbw >= 1 then
			JABOB_Settings.bbw = JABOB_Settings.bbw -1
			JB.ResizeBag("bank")
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")
	
	_ui.sbcontrols.spacing_v = UI.CreateFrame("Texture", "_ui.sbcontrols.spacing_v", _ui.sbcontrols.config)
	_ui.sbcontrols.spacing_v:SetPoint("TOPLEFT", _ui.sbcontrols.spacing_h, "TOPRIGHT", 10, 0)
	_ui.sbcontrols.spacing_v:SetTexture(addon.identifier, "textures/space_v.png")
	_ui.sbcontrols.spacing_v:SetLayer(12)
	_ui.sbcontrols.spacing_v:SetWidth(32)
	_ui.sbcontrols.spacing_v:SetHeight(32)
	_ui.sbcontrols.spacing_v:SetVisible(true)

	_ui.sbcontrols.spacing_v:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if JABOB_Settings.bsy <= 7 then
			JABOB_Settings.bsy = JABOB_Settings.bsy +1
			JB.ArrangeBagSlots("bank")
		end
	end, "Event.UI.Input.Mouse.Wheel.Forward")

	_ui.sbcontrols.spacing_v:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if JABOB_Settings.bsy >= 1 then
			JABOB_Settings.bsy = JABOB_Settings.bsy -1
			JB.ArrangeBagSlots("bank")
		end
	end, "Event.UI.Input.Mouse.Wheel.Back")	
	
	for sk, sv in pairs(Library.LibInventory.REFSLOTS["bank"]) do
		WATCHDOGWAIT(STRINGFORMAT("BuildUI:4:%s", sk))
		_ui.bagslots[sk] = JB.BuildBagElement(sk, _ui.playerbnk)
	end

	for sk, sv in pairs(Library.LibInventory.REFSLOTS["vault"]) do
		WATCHDOGWAIT(STRINGFORMAT("BuildUI:4:%s", sk))
		_ui.bagslots[sk] = JB.BuildBagElement(sk, _ui.playerbnk)
	end
	
	WATCHDOGWAIT("BuildUI:6")

	_ui.tooltip.t1.frame = UI.CreateFrame("Frame", "_ui.tooltip.t1.frame", _ui.context)
	_ui.tooltip.t1.frame:SetVisible(false)
	_ui.tooltip.t1.frame:SetLayer(100)
	_ui.tooltip.t1.frame:SetBackgroundColor(0,0,0)
	
	_ui.tooltip.t1.count = UI.CreateFrame("Frame", "_ui.tooltip.t1.count", _ui.tooltip.t1.frame)
	_ui.tooltip.t1.count:SetPoint("BOTTOMLEFT", _ui.tooltip.t1.frame, "TOPLEFT", 0, -20)
	_ui.tooltip.t1.count:SetPoint("BOTTOMRIGHT", _ui.tooltip.t1.frame, "TOPRIGHT", 0, -20)
	_ui.tooltip.t1.count:SetVisible(false)
	_ui.tooltip.t1.count:SetBackgroundColor(0,0,0,0.5)
	
	_ui.tooltip.t1.count.details = {}
	
	_ui.tooltip.c1.frame = UI.CreateFrame("Frame", "_ui.tooltip.c1.frame", _ui.context)
	_ui.tooltip.c1.frame:SetVisible(false)
	_ui.tooltip.c1.frame:SetLayer(100)
	_ui.tooltip.c1.frame:SetBackgroundColor(0,0,0)

	_ui.tooltip.c2.frame = UI.CreateFrame("Frame", "_ui.tooltip.c2.frame", _ui.context)
	_ui.tooltip.c2.frame:SetVisible(false)
	_ui.tooltip.c2.frame:SetLayer(100)
	_ui.tooltip.c2.frame:SetBackgroundColor(0,0,0)

	for x=1, 3 do
		local f
		if x == 1 then
			f = _ui.tooltip.t1
		elseif x == 2 then
			f = _ui.tooltip.c1
		else
			f = _ui.tooltip.c2
		end

		f.borders = {}

		f.borders.tl = UI.CreateFrame("Texture", "f.borders.tl", f.frame)
		f.borders.tl:SetTexture(addon.identifier, "textures/border_tl.png")
		f.borders.tl:SetPoint("BOTTOMRIGHT", f.frame, "TOPLEFT",5,5)
		f.borders.tl:SetLayer(95)

		f.borders.tr = UI.CreateFrame("Texture", "f.borders.tr", f.frame)
		f.borders.tr:SetTexture(addon.identifier, "textures/border_tr.png")
		f.borders.tr:SetPoint("BOTTOMLEFT", f.frame, "TOPRIGHT",-5,5)
		f.borders.tr:SetLayer(95)

		f.borders.bl = UI.CreateFrame("Texture", "f.borders.bl", f.frame)
		f.borders.bl:SetTexture(addon.identifier, "textures/border_bl.png")
		f.borders.bl:SetPoint("TOPRIGHT", f.frame, "BOTTOMLEFT",5,-5)
		f.borders.bl:SetLayer(95)

		f.borders.br = UI.CreateFrame("Texture", "f.borders.br", f.frame)
		f.borders.br:SetTexture(addon.identifier, "textures/border_br.png")
		f.borders.br:SetPoint("TOPLEFT", f.frame, "BOTTOMRIGHT",-5,-5)
		f.borders.br:SetLayer(95)

		f.borders.t = UI.CreateFrame("Texture", "f.borders.t", f.frame)
		f.borders.t:SetTexture(addon.identifier, "textures/border_t.png")
		f.borders.t:SetPoint("TOPLEFT", f.borders.tl, "TOPRIGHT",0,0)
		f.borders.t:SetLayer(90)

		f.borders.l = UI.CreateFrame("Texture", "f.borders.l", f.frame)
		f.borders.l:SetTexture(addon.identifier, "textures/border_l.png")
		f.borders.l:SetPoint("TOPLEFT", f.borders.tl, "BOTTOMLEFT",0,0)
		f.borders.l:SetLayer(90)

		f.borders.b = UI.CreateFrame("Texture", "f.borders.b", f.frame)
		f.borders.b:SetTexture(addon.identifier, "textures/border_b.png")
		f.borders.b:SetPoint("BOTTOMLEFT", f.borders.bl, "BOTTOMRIGHT",0,0)
		f.borders.b:SetLayer(90)

		f.borders.r = UI.CreateFrame("Texture", "f.borders.r", f.frame)
		f.borders.r:SetTexture(addon.identifier, "textures/border_r.png")
		f.borders.r:SetPoint("TOPRIGHT", f.borders.tr, "BOTTOMRIGHT",0,0)
		f.borders.r:SetLayer(90)
	end

	for x=1, 40 do
		WATCHDOGWAIT("BuildUI:7:"..x)
		local f1 = {}
		local f2 = {}
		local f3 = {}
		f1.info = UI.CreateFrame("Text", "f1.info", _ui.tooltip.t1.frame)
		f1.info:SetLayer(101)
		f1.info:SetText(tostring(x))
		f2.info = UI.CreateFrame("Text", "f2.info", _ui.tooltip.c1.frame)
		f2.info:SetLayer(101)
		f2.info:SetText(tostring(x))
		f3.info = UI.CreateFrame("Text", "f3.info", _ui.tooltip.c2.frame)
		f3.info:SetLayer(101)
		f3.info:SetText(tostring(x))

		if x == 1 then
			f1.info:SetFontSize(16)
			f2.info:SetFontSize(12)
			f3.info:SetFontSize(12)
			_ui.h16 = f1.info:GetHeight()
			f1.info:SetPoint("TOPLEFT", _ui.tooltip.t1.frame, "TOPLEFT", 0,0)
			f2.info:SetPoint("TOPLEFT", _ui.tooltip.c1.frame, "TOPLEFT", 0,0)
			f3.info:SetPoint("TOPLEFT", _ui.tooltip.c2.frame, "TOPLEFT", 0,0)
		else
			f1.info:SetFontSize(12)
			if x == 2 then
				_ui.h12 = f1.info:GetHeight()
				f2.info:SetFontSize(16)
				f3.info:SetFontSize(16)
			else
				f2.info:SetFontSize(12)
				f3.info:SetFontSize(12)
			end
			f1.info:SetPoint("TOPLEFT", _ui.tooltip.t1.lines[x-1].info, "BOTTOMLEFT", 0,0)
			f2.info:SetPoint("TOPLEFT", _ui.tooltip.c1.lines[x-1].info, "BOTTOMLEFT", 0,0)
			f3.info:SetPoint("TOPLEFT", _ui.tooltip.c2.lines[x-1].info, "BOTTOMLEFT", 0,0)
		end
		_ui.tooltip.t1.lines[x] = f1
		_ui.tooltip.c1.lines[x] = f2
		_ui.tooltip.c2.lines[x] = f3
	end
end

function JB.CloseBags(bt)
	if bt == "inventory" then
		_invbagsvisible = false
		_ui.playerinv:SetVisible(false)
		_ui.sicontrols.search:SetText("")
		_ui.sicontrols.search:SetKeyFocus(false)
		JB.SearchBags("inventory")
	elseif bt == "bank" then
		_ui.playerbnk:SetVisible(false)
		_ui.sbcontrols.search:SetText("")
		_ui.sbcontrols.search:SetKeyFocus(false)
		JB.SearchBags("bank")
	end
	_ui.tooltip.t1.frame:SetVisible(false)
	_ui.tooltip.c1.frame:SetVisible(false)
	_ui.tooltip.c2.frame:SetVisible(false)
	popupCur = ""
	CMDTOOLTIP = nil
	for k,v in pairs(SLOT_PULSE) do
		_ui.bagslots[k].frame:SetVisible(false)
	end
	SLOT_PULSE = {}
	SLOT_PULSE_ACTIVE = false
end

local curalpha = 1
local pulsetime = 0
local pulsecnt = 0

local pulse_img = {
	"textures/rot1.png",
	"textures/rot2.png",
	"textures/rot3.png",
	"textures/rot4.png",
	"textures/rot5.png",
	"textures/rot6.png",
	"textures/rot7.png",
	"textures/rot8.png"
}

function JB.PulseSlots()
	local tx = Inspect.Time.Real()
	if tx-pulsetime > 0.1 then
		pulsecnt = (pulsecnt%8)+1
		for k,v in pairs(SLOT_PULSE) do
			_ui.bagslots[k].frame:SetTexture(addon.identifier, pulse_img[pulsecnt])
		end
		pulsetime = tx
	end
end

function JB.xPulseSlots()
	local tx = Inspect.Time.Real()
	if tx-pulsetime > 0.1 then
		curalpha = curalpha + sp_inc
		for k,v in pairs(SLOT_PULSE) do
			_ui.bagslots[k].frame:SetAlpha(curalpha)
		end
		if curalpha <= 0.1 then
			sp_inc = 0.1
		elseif curalpha >= 1 then
			sp_inc = -0.1
		end
		pulsetime = tx
	end
end

function JB.cb_libinvBAG(t)
	local refresh = {
		["inventory"] = false,
		["bank"] = false
		}
	for k,v in pairs(t) do
		local ix, ty, sl = Utility.Item.Slot.Parse(k)
		if v and v ~= "nil" then
			if szBags[ix][sl] ~= v.slots then
				szBags[ix][sl] = v.slots
				refresh[ix] = true
			end		
		else
			if szBags[ix][sl] ~= 0 then
				szBags[ix][sl] = 0
				refresh[ix] = true
			end
		end
	end
	_bagupdate.inventory.arrange = refresh["inventory"]
	_bagupdate.bank.arrange = refresh["bank"]
end

local initialBagSetupDone = false

function JB.cb_libinvSLOT(t)
	if UINEEDED == false then
		for si,d in pairs(t) do
			--print("A> "..si.." : "..tostring(d).." : "..type(d))
			local prevID = _SLOTID[si]
			local ix, ty, sl = Utility.Item.Slot.Parse(si)
			if d and d ~= "nil" then
				_ui.bagslots[si].icon:SetTexture("Rift", d.icon)
				_ui.bagslots[si].icon:SetVisible(true)
				if d.stack ~= nil and d.stack > 1 then
					_ui.bagslots[si].text:SetText(STRINGFORMAT("%d", d.stack))
					_ui.bagslots[si].text:SetVisible(true)
				else
					_ui.bagslots[si].text:SetVisible(false)
				end
				local cd = getRarityRGB(d.rarity)
				_ui.bagslots[si].container:SetBackgroundColor(cd.r, cd.g, cd.b)
				_ui.bagslots[si].frame:SetVisible(false)
				SLOT_PULSE[si] = nil
				_SLOTID[si] = d.id
				_SLOTSINUSE[ix][si] = d
				if prevID and prevID ~= d.id and CDSLOTS[si] then
					CDSLOTS[si] = nil
					prevCDUpdate = 0
					_ui.bagslots[si].cd:SetVisible(false)
				end
				if initialBagSetupDone and ix == "inventory" and _ui.playerinv:GetVisible() == false and prevID ~= d.id then
					_ui.bagslots[si].frame:SetVisible(true)
					SLOT_PULSE[si] = true
					SLOT_PULSE_ACTIVE = true
				end
			else
				_ui.bagslots[si].icon:SetVisible(false)
				_ui.bagslots[si].text:SetVisible(false)
				_ui.bagslots[si].container:SetBackgroundColor(0,0,0,0.5)
				_ui.bagslots[si].container:SetAlpha(1)
				_SLOTID[si] = nil
				_SLOTSINUSE[ix][si] = nil
				if SLOT_PULSE[si] then
					SLOT_PULSE[si] = nil
					_ui.bagslots[si].frame:SetVisible(false)
				end
				if CDSLOTS[si] then
					CDSLOTS[si] = nil
					prevCDUpdate = 0
					_ui.bagslots[si].cd:SetVisible(false)
				end
			end
		end
		if initialBagSetupDone then
			JB.SearchBags("inventory")
			JB.SearchBags("bank")
		end
		initialBagSetupDone = true
	else
		--print("Delaying cb handle of libinvSLOT")
		table.insert(data_libinvSLOT, t)
	end
end

function JB.UpdateCooldowns()
	local tx = Inspect.Time.Real()
	if tx-prevCDUpdate > 0.2 then
		CDACTIVE = false
		for k,v in pairs(CDSLOTS) do
			if Library.LibInventory.REFSLOTS["inventory"][k] then
				if tx < v then
					CDACTIVE = true
					local dx = math.floor(v-tx)
					if dx >= 3600 then
						_ui.bagslots[k].cd:SetText(">1hr")
					elseif dx > 300 then
						_ui.bagslots[k].cd:SetText(STRINGFORMAT("%dm", dx/60))
					else
						local m = math.floor(dx/60)
						local s = dx-(m*60)
						_ui.bagslots[k].cd:SetText(STRINGFORMAT("%d:%02d", m,s))
					end
					_ui.bagslots[k].cd:SetVisible(true)
				else
					CDSLOTS[k] = nil
					_ui.bagslots[k].cd:SetVisible(false)
				end
			else
				CDSLOTS[k] = nil
			end
		end
		prevCDUpdate = tx
	end
end

function JB.cb_libinvCD(t)
	for k,v in pairs(t) do
		for s,_ in pairs(v.slots) do
			CDSLOTS[s] = v.endtime
			--print(string.format("%s [%s] = %.4f", s, k, v.endtime))
		end
	end
	CDACTIVE = true
end

function JB.Event_Unit_Availability_Full(h,t)
	for k,v in pairs(t) do
		if v == "player" then
			Debug("JB.Event_Unit_Availability_Full(player)", INFORMATION)
			Command.Event.Detach(Event.Unit.Availability.Full, nil, nil, nil, addon.identifier)
			
			LibVersionCheck.register(addon.toc.Identifier, addon.toc.Version)
			
			Library.LibInventory.RegisterSlot(Library.LibInventory.REFSLOTS["bankbag"], JB.cb_libinvBAG)
			Library.LibInventory.RegisterSlot(Library.LibInventory.REFSLOTS["inventorybag"], JB.cb_libinvBAG)
			Library.LibInventory.RegisterSlot(Library.LibInventory.REFSLOTS["bank"], JB.cb_libinvSLOT)
			Library.LibInventory.RegisterSlot(Library.LibInventory.REFSLOTS["vault"], JB.cb_libinvSLOT)
			Library.LibInventory.RegisterSlot(Library.LibInventory.REFSLOTS["inventory"], JB.cb_libinvSLOT)
			Library.LibInventory.Cooldowns(JB.cb_libinvCD)
		end
	end
end

function JB.Command_Slash_Register(h, args)
	local r = {}
	local numargs = 0
	for token in string.gmatch(args, "[^%s]+") do
		numargs=numargs+1
		r[numargs] = token
	end

	if r[1] == "toggle" then
		_invbagsvisible = not _invbagsvisible
		if _invbagsvisible == false then
			JB.CloseBags("inventory")
			JB.CloseBags("bank")
		else
			_ui.playerinv:SetVisible(_invbagsvisible)
			if Inspect.Interaction("bank") then
				_ui.playerbnk:SetVisible(_invbagsvisible)
			else
				_ui.playerbnk:SetVisible(false)
			end
		end
	elseif r[1] == "tooltip" then
		if JABOB_Settings.tooltip == JB_TOOLTIP_ALL then
			JABOB_Settings.tooltip = JB_TOOLTIP_RFT
			Debug("Using Rift API tooltip", RELEASE)
		elseif JABOB_Settings.tooltip == JB_TOOLTIP_RFT then
			JABOB_Settings.tooltip = JB_TOOLTIP_CUS
			Debug("Using JABOB tooltips", RELEASE)
		else
			JABOB_Settings.tooltip = JB_TOOLTIP_ALL
			Debug("Using Rift API & JABOB tooltips", RELEASE)
		end
	elseif r[1] == "border" then
		local dx = tonumber(r[2])
		if dx then dx = math.abs(dx) end
		if dx >=0 and dx <=5 then
			JABOB_Settings.bw = dx
		end
		JB.ResizeBag("inventory")
		JB.ResizeBag("bank")
	elseif r[1] == "space" then
		local dx = tonumber(r[2])
		if dx then dx = math.abs(dx) end
		if dx >=0 and dx <=5 then
			JABOB_Settings.sx = dx
			JABOB_Settings.sy = dx
		end
		JB.ArrangeBagSlots("inventory")
		JB.ArrangeBagSlots("bank")
	elseif r[1] == "spacex" then
		local dx = tonumber(r[2])
		if dx then dx = math.abs(dx) end
		if dx >=0 and dx <=5 then
			JABOB_Settings.sx = dx
		end
		JB.ArrangeBagSlots("inventory")
		JB.ArrangeBagSlots("bank")
	elseif r[1] == "spacey" then
		local dx = tonumber(r[2])
		if dx then dx = math.abs(dx) end
		if dx >=0 and dx <=5 then
			JABOB_Settings.sy = dx
		end
		JB.ArrangeBagSlots("inventory")
		JB.ArrangeBagSlots("bank")
	end
end

function JB.UpdateHandler()
	while true do
		if UINEEDED then
			JB.BuildUI()
			UINEEDED = false
		elseif SVLOADED then
			JB.procEvent_Addon_SavedVariables_Load_End()
			SVLOADED = false
		elseif #data_libinvSLOT > 0 then
			--print("processing catch up event data..")
			JB.cb_libinvSLOT(data_libinvSLOT[1])
			table.remove(data_libinvSLOT,1)
		else
			if _bagupdate.inventory.resize then
				JB.doResizeBag("inventory")
				_bagupdate.inventory.resize = false
			end
			if _bagupdate.inventory.arrange then
				JB.doArrangeBagSlots("inventory")
				_bagupdate.inventory.arrange = false
			end
			if _bagupdate.bank.resize then
				JB.doResizeBag("bank")
				_bagupdate.bank.resize = false
			end
			if _bagupdate.bank.arrange then
				JB.doArrangeBagSlots("bank")
				_bagupdate.bank.arrange = false
			end
			if SLOT_PULSE_ACTIVE and _invbagsvisible then
				JB.PulseSlots()
			end
			if CDACTIVE and _invbagsvisible then
				JB.UpdateCooldowns()
			end
		end
		coroutine.yield()
	end
end

JB.coUpdateHandler = coroutine.create(JB.UpdateHandler)

function JB.Event_System_Update_Begin(h)
	if CMDTOOLTIP ~= false then
		Command.Tooltip(CMDTOOLTIP)
		CMDTOOLTIP = false
	elseif CONFIG_INV == true then
		local md = Inspect.Mouse()
		local t = math.min(_ui.sicontrols.options:GetTop(), _ui.sicontrols.config:GetTop())
		local b = math.max(_ui.sicontrols.options:GetBottom(), _ui.sicontrols.config:GetBottom())
		local l = math.min(_ui.sicontrols.options:GetLeft(), _ui.sicontrols.config:GetLeft())
		local r = math.max(_ui.sicontrols.options:GetRight(), _ui.sicontrols.config:GetRight())
		if md.x < l or md.x > r or md.y < t or md.y > b then
			_ui.sicontrols.config:SetVisible(false)
			CONFIG_INV = false
		end
	elseif CONFIG_BNK == true then
		local md = Inspect.Mouse()
		local t = math.min(_ui.sbcontrols.options:GetTop(), _ui.sbcontrols.config:GetTop())
		local b = math.max(_ui.sbcontrols.options:GetBottom(), _ui.sbcontrols.config:GetBottom())
		local l = math.min(_ui.sbcontrols.options:GetLeft(), _ui.sbcontrols.config:GetLeft())
		local r = math.max(_ui.sbcontrols.options:GetRight(), _ui.sbcontrols.config:GetRight())
		if md.x < l or md.x > r or md.y < t or md.y > b then
			_ui.sbcontrols.config:SetVisible(false)
			CONFIG_BNK = false
		end
	end

	if coroDead then return end
	local ok, errormsg = coroutine.resume(JB.coUpdateHandler)
	if not ok then
		coroDead = true
		error("Error in coroutine: "..errormsg)
	end
end

function JB.Event_Addon_SavedVariables_Load_End(h, a)
	if a == addon.identifier then
		SVLOADED = true
	end
end

---------------------
-- Register Events
---------------------

Command.Event.Attach(Event.Addon.SavedVariables.Load.End, JB.Event_Addon_SavedVariables_Load_End, "Event.Addon.SavedVariables.Load.End")
Command.Event.Attach(Command.Slash.Register("jabob"), JB.Command_Slash_Register, "Command.Slash.Register")
Command.Event.Attach(Event.System.Update.Begin, JB.Event_System_Update_Begin, "Event.System.Update.Begin")
Command.Event.Attach(Event.Unit.Availability.Full, JB.Event_Unit_Availability_Full, "Event.Unit.Availability.Full")

print(STRINGFORMAT("v%s loaded.", addon.toc.Version))